"""Robotmk CLI Interface. 

Start Robotmk in different contexts.
Context command can also be set via environment variable ROBOTMK_common_context."""
import click
from robotmk.main import Robotmk, DEFAULTS, LOG_LEVELS
import importlib
import pkgutil
import os.path
import warnings


# CMD1        CMD2     OPTION                                    CMD3            # Description
# ---------------------------------------------------------------------------------------------
# # FS CONTEXT
# robotmk                                                                        # no arg = print output

# robotmk     fs                                              output          # print output
# robotmk     fs       --yml /etc/checkmk/another_robotmk.yml    output          # print output with yml
# robotmk     fs       --vars /var/robotmk_local.env             output          # print output, load env from file (instead of env)

# robotmk     fs                                                 scheduler       # start scheduler
# robotmk     fs       --yml /etc/checkmk/another_robotmk.yml    scheduler       # start scheduler with yml
# robotmk     fs       --vars /var/robotmk_local.env             scheduler       # start scheduler, load env from file  (instead of env)
# ---------------------------------------------------------------------------------------------
# # SUITE CONTEXT
# robotmk                                                                        # no arg = exec suite as configured in env
# robotmk     suite    --vars /var/rmk/foosuiteA_8bb36c3.env                     # exec suite with env from file and suite = yml -> common: suite)
# robotmk     suite    --vars /var/rmk/foosuiteA_8bb36c3.env     bazsuite        # exec suite with env from file and suite = bazsuite
# robotmk     suite                                              vardump  foobarsuiteA   # just dump the vars for foobarsuiteA
# ---------------------------------------------------------------------------------------------
# # SPECIALAGENT (="s.a."") CONTEXT
# robotmk                                                                        # no arg = seq & output
# (robotmk    s.a.     output)                                                   # NOT POSSIBLE - no config file
# robotmk     s.a.     --vars ~/var/robotmk/s.a.-hostfoo.env     output          # run output with env from file

# (robotmk    s.a.     sequencer)                                                # NOT POSSIBLE - no config file
# robotmk     s.a.     --vars ~/var/robotmk/s.a.-hostfoo.env     sequencer       # run requencer with env from file


def get_commands_from_pkg(pkg) -> dict:
    """CLI options for the contexts are defined within the context packages within cli.py module.

    Each sub-PACKAGE of pkg is considered a sub-COMMAND. This function recursively traverses the
    package tree and returns all function objects which have the same name as the package.

    This allows to maintain the CLI logic within the context packages, while still having a single
    entry point for the CLI.

    Example:
    robotmk.context.agent.cli.py contains a function called agent(), decorated with @click.group()
    and connected with subcomands "output()" and "scheduler()". The discovered subcommand is "agent()".
    """
    pkg_obj = importlib.import_module(pkg)
    pkg_path = os.path.dirname(pkg_obj.__file__)
    commands = {}
    for module in pkgutil.iter_modules([pkg_path]):
        module_obj = importlib.import_module(f"{pkg}.{module.name}")
        if module.ispkg:
            cmd_from_pkg = get_commands_from_pkg(f"{pkg}.{module.name}")
            commands.update(cmd_from_pkg)
        else:
            if module.name == "cli":
                cli_functions = [f for f in dir(module_obj) if f == pkg.split(".")[-1]]

                for cli_function in cli_functions:
                    commands[cli_function] = getattr(module_obj, cli_function)

    return commands


# Create the main group and assign the subcommands gathered from the context packages
@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    help=__doc__,
    invoke_without_command=True,
    commands=get_commands_from_pkg("robotmk.context"),
)
@click.option(
    "--loglevel",
    "-l",
    default=DEFAULTS["common"]["log_level"],
    type=click.Choice(LOG_LEVELS),
)
@click.pass_context
def main(ctx, loglevel):
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@main.command()
def diagnose():
    """Print diagnostic information."""
    click.secho("diagnose", fg="yellow")
    # TODO implement diagnose()
    # - environment with ROBOTOMK vars
    # YML present?
    # RCC present
    # OS
    # Python version
    # Robotmk version
    # Robot version
    # configdump
    # create a dummy suite with api and run it in suite mode
    # CPU cores
    # RAM
    # disk space
    # network

    pass


if __name__ == "__main__":
    main()



import os
import math

from robotmk.logger import RobotmkLogger
from .abstract import LocalTarget
from ...strategies import RunStrategy


# TODO: make RCC binary path configurabe


class RCCTarget(LocalTarget):
    def __init__(
        self,
        suiteuname: str,
        config: dict,
        logger: RobotmkLogger,
    ):
        super().__init__(suiteuname, config, logger)

    @staticmethod
    def worker_count():
        """Determines the number of RCC worker processes by the total CPU cores available.

        Example:
        - 4 cores: 2 workers
        - 5 cores: 2 workers
        - 6 cores: 3 workers
        - 7 cores: 3 workers
        - 8 cores: 4 workers
        """
        return max(min(math.floor(os.cpu_count() / 2), 2), 6)

    def __str__(self) -> str:
        return "rcc"

    @property
    def pre_command(self) -> list:
        """Runs just before the command, executed by the Run strategy"""
        return [
            "rcc",
            "holotree",
            "vars",
            "--controller",
            "robotmk",
            "--space",
            self.suiteuname,
            "--workers",
            str(self.worker_count()),
            "-r",
            str(self.path.joinpath("robot.yaml")),
        ]

    @property
    def main_command(self) -> list:
        """The command will be used by the Run strategy (self.target.command).

        In RCC target, the commandline gets buuilt to execute a RCC task (=Robotmk inside of RCC)
        """
        # DEBUG: "; ".join([f"export {k}={v}" for (k,v) in kwargs.get("env").items() if k.startswith("RO")])
        # paste export strings into shell
        # - which python3 -> shout point to RCC venv
        # - env -> ROBOCORP and ROBOTMK vars should be set
        # - which robotmk -> path to entrypoint
        # python3 -m trace --trace <ENTRYPOINT> suite run suite_default
        return [
            "rcc",
            "task",
            "run",
            "--controller",
            "robotmk",
            "--space",
            self.suiteuname,
            "-t",
            "robotmk",
            "-r",
            str(self.path.joinpath("robot.yaml")),
        ]

    def prepare_environment(self) -> dict:
        """Sets up the environment for a subsequent RCC robot run.

        If Robotmk calls itself in a RCC task, the inner call of Robotmk needs
        some special settings, e.g. NOT to use RCC again, to log into the default
        logdir, etc.
        This function first exports the current config to the environment and
        then adds the special settings on top.
        """
        env = os.environ.copy()
        added_settings = {"suitecfg.run.rcc": False, "suitecfg.uuid": self.uuid}
        self.config.cfg_to_environment(self.config.configdict, environ=env)
        self.config.dotcfg_to_env(added_settings, environ=env)
        return env

    def run(self):
        # Before we blow up the whole thing, we should check if the suite is RCC compatible and allowed to run
        if self.is_disabled_by_flagfile:
            # TODO: Log skipped
            # reason = self.get_disabled_reason()
            return
        if not self.is_rcc_compatible:
            # TODO: log suite not compatible with RCC
            pass
        else:
            run_env = self.prepare_environment()
            self.rc = self.run_strategy.run(env=run_env)

    @property
    def is_rcc_compatible(self):
        """Returns True if the given suite folder is compatible with RCC.
        Such a suite dir must at least contain conda.yml and robot.yml.
        """
        if (
            self.path.joinpath("conda.yaml").exists()
            and self.path.joinpath("robot.yaml").exists()
        ):
            return True
        else:
            return False

    # def calculate_blueprint_hash(self):
    #     try:
    #         output = subprocess.check_output(
    #             ["rcc", "ht", "hash", self.blueprint_path], universal_newlines=True
    #         )
    #         self.blueprint_hash = output.strip()
    #     except subprocess.CalledProcessError as e:
    #         raise RuntimeError(
    #             f"Failed to calculate blueprint hash: {e.stderr.strip()}"
    #         )

    # def is_environment_ready(self):
    #     if not self.blueprint_hash:
    #         self.calculate_blueprint_hash()
    #     try:
    #         output = subprocess.check_output(
    #             ["rcc", "ht", "spaces", "--filter", self.blueprint_hash],
    #             universal_newlines=True,
    #         )
    #         spaces = json.loads(output)
    #         return bool(spaces)
    #     except subprocess.CalledProcessError as e:
    #         raise RuntimeError(
    #             f"Failed to check environment readiness: {e.stderr.strip()}"
    #         )

    # def check_spaces(self):
    #     if not self.blueprint_hash:
    #         self.calculate_blueprint_hash()
    #     try:
    #         output = subprocess.check_output(
    #             ["rcc", "ht", "spaces", "--filter", self.blueprint_hash],
    #             universal_newlines=True,
    #         )
    #         spaces = json.loads(output)
    #         return spaces
    #     except subprocess.CalledProcessError as e:
    #         raise RuntimeError(f"Failed to check spaces: {e.stderr.strip()}")

    # def create_environment(self, name, variables=None):
    #     if not self.blueprint_hash:
    #         self.calculate_blueprint_hash()
    #     cmd = ["rcc", "ht", "vars", "--blueprint", self.blueprint_hash, "--name", name]
    #     if variables:
    #         cmd.extend(["--vars", json.dumps(variables)])
    #     try:
    #         subprocess.check_call(cmd)
    #     except subprocess.CalledProcessError as e:
    #         raise RuntimeError(f"Failed to create environment: {e.stderr.strip()}")



class ContextFactory:
    def __init__(self, contextname: str, log_level: str) -> None:
        self.contextname = contextname
        self._log_level = log_level

    def get_context(self) -> None:
        if self.contextname == "agent":
            from .agent.agent import AgentContext

            return AgentContext()
        elif self.contextname == "specialagent":
            from .specialagent.specialagent import SpecialAgentContext

            return SpecialAgentContext()
        elif self.contextname == "suite":
            from .suite.suite import SuiteContext

            return SuiteContext()
        else:
            # TODO: catch this error
            pass



from pathlib import Path
import mergedeep

from robotmk.logger import RobotmkLogger

from .retry import RetryStrategyFactory
from .state import RFState
from ..abstract import LocalTarget


from datetime import datetime

local_tz = datetime.utcnow().astimezone().tzinfo

# This is the "heavy" part of the code. It contains the logic to run RF.
# Here I place all the Robotmk v1 shit.


class RobotFrameworkTarget(LocalTarget):
    def __init__(
        self,
        suiteuname: str,
        config: dict,
        logger: RobotmkLogger,
    ):
        super().__init__(suiteuname, config, logger)

    def __str__(self) -> str:
        return "robotframework"

    @property
    def main_command(self) -> list:
        """The command will be used by the Run strategy (self.target.command).

        In RF target, the complete commandline must be built to execute the RF suite.
        (See https://robot-framework.readthedocs.io/en/latest/autodoc/robot.html#robot.run.run_cli)
        TODO: Logging
        """
        self.robot_params.update(
            {
                "log": self.log_html,  #'rf_suite_default_1682429039_10559086-1.html'
                "output": self.output_xml,  #'rf_suite_default_1682429039_10559086-1.xml'
            }
        )

        suite_params = mergedeep.merge(
            {}, self.config.get("suitecfg.params").asdict(), self.robot_params
        )
        arglist = ["robot"]
        for k, v in suite_params.items():
            arg = f"--{k}"
            # create something we can iterate over
            if isinstance(v, str):
                # key:value    => convert to 1 el list
                vlist = [v]
            elif isinstance(v, dict):
                if k == "variable":
                    # key:var-dict => convert to list of varkey:varvalue
                    vlist = list(map(lambda x: f"{x[0]}:{x[1]}", v.items()))
                else:
                    self._suite.logger.warn(
                        f"The Robot Framework parameter {k} is a dict but cannot be converted to cmdline arguments (values: {str(v)})"
                    )
            elif isinstance(v, list):
                if k == "argumentfile" or k == "variablefile":
                    # make the file args absolute file paths
                    v = [str(self._suite.pathdir.joinpath(n)) for n in v]
                # key:list     => no conversion
                vlist = v

            for value in vlist:
                # values which are boolean(-like) are single parameters without option
                if type(value) is bool or value in ["yes", "no", "True", "False"]:
                    arglist.extend([arg])
                else:
                    arglist.extend([arg, value])
        # the path of the robot suite is the very last argument
        arglist.append(str(self.path))
        return arglist

    def run(self):
        self.retry_strategy = RetryStrategyFactory(self).create()

        self.shortuuid = self.uuid[:8]
        # this timestamp is used to keep all result files in order; it is used
        # for all target executions
        self._timestamp = self.get_now_as_epoch()
        self._state = RFState(self)
        # params for RF: global ones & re-execution
        # self.robot_params = {"console": "NONE", "report": "NONE"}
        self.robot_params = {"report": "NONE"}

        # Do not run if the suite dir contains a DISABLED file
        if self.is_disabled_by_flagfile:
            # TODO: Log if a suite was skipped
            # reason = self.get_disabled_reason()
            return
        else:
            # RF "outputdir" = logdir + robotframework (store HTML, XML and console logs)
            self.robot_params.update({"outputdir": self.outputdir})
            self._state.timer_start()
            self.rc = self.retry_strategy.run()
            self._state.timer_stop()
            self._state.write()
            pass

    def _get_now_as_dt(self):
        return datetime.now(local_tz)

    def get_now_as_epoch(self):
        return int(self._get_now_as_dt().timestamp())

    @property
    def timestamp(self):
        """Returns the timestamp the suite execution was started. This is
        used for all executions of the suite, including retries in order
        to group the result files."""
        return self._timestamp

    def get_disabled_reason(self) -> str:
        """Report back the reason why the suite was disabled."""
        if self.is_disabled_by_flagfile:
            try:
                with open(self.path.joinpath("DISABLED"), "r") as f:
                    reason = f.read()
                    if len(reason) > 0:
                        return "Reason: " + reason
                    else:
                        return ""
            except:
                return ""

    # OUTPUT

    @property
    def outputdir(self):
        return str(Path(self.logdir).joinpath("robotframework"))

    @property
    def output_filename(self):
        """Returns the output filename string, including the retry number.

        Example:
            rf_suite1_978741fb_1680335851
            rf_suite1_978741fb_1680335851-1"""
        if self.attempt is None:
            suite_filename = "rf_%s_%s_%s" % (
                self.suiteuname,
                self.timestamp,
                self.shortuuid,
            )
        else:
            suite_filename = "rf_%s_%s_%s-%d" % (
                self.suiteuname,
                self.timestamp,
                self.shortuuid,
                int(self.attempt),
            )
        return suite_filename

    # XML Output---
    @property
    def output_xml(self):
        return self.output_filename + ".xml"

    @property
    def output_xml_fullpath(self):
        return str(Path(self.outputdir).joinpath(self.output_xml))

    # HTML Output ---
    @property
    def log_html(self):
        return self.output_filename + ".html"

    @property
    def log_html_fullpath(self):
        return str(Path(self.outputdir).joinpath(self.log_html))

    # Suite timestamp for filenames
    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, t):
        self._timestamp = t



import yaml
from robotmk.main import Robotmk
import os

cwd = os.path.dirname(__file__)
robotmk_yml = os.path.join(cwd, "robotmk.yml")
robotmk_env = os.path.join(cwd, "robotmk.env")


def test_suite_context():
    """Tests if a Robotmk object can be created for suite context."""
    robotmk = Robotmk(contextname="suite", log_level=None, ymlfile=robotmk_yml)
    assert robotmk.config.get("common.suiteuname") == "suite_ospython"



"""Robotmk main module

The main module of the Robotmk package, where the Robotmk instance gets
created for the use within the contexts.
It also contains the DEFAULTS dict with all default values for the config. """


import os
import sys
from loguru import logger
from robotmk.context import ContextFactory

# TODOs:
# - add logging
# - add pytests

LOG_LEVELS = ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]

DEFAULTS = {
    # Default values for the "common" config section
    "common": {
        "robotmk_yml": "robotmk.yml",
        "log_level": "INFO",
        "log_retention": 14,
    },
    # Default values for the "common" config section (Windows)
    "nt": {
        "cfgdir": "C:/ProgramData/checkmk/agent/config",  # where to find robotmk.yml
        "robotdir": "C:/ProgramData/checkmk/agent/robot",
        "logdir": "C:/ProgramData/checkmk/agent/log/robotmk/",
        "tmpdir": "C:/ProgramData/checkmk/agent/tmp/robotmk",
    },
    # Default values for the "common" config section (Linux)
    "posix": {
        "cfgdir": "/etc/check_mk",  # where to find robotmk.yml
        "robotdir": "/usr/lib/check_mk_agent/robot",
        "logdir": "/var/log/robotmk",
        "tmpdir": "/tmp/robotmk",
    },
}


class Robotmk:
    """This is the main class of the robotmk package. It is used to create a
    Robotmk instance with a specific context.

    Configuration loading (yml, var, env) depends on the context and is only
    done when no config object is passed to the init method."""

    def __init__(
        self,
        log_level=None,
        contextname=None,
        ymlfile: str = None,
        varfile: str = None,
        default_cfg: dict = {},
    ) -> None:
        """context is the strategy to use and in fact a set of factory methods.
        If called from the CLI without context argument, the default context
        will be read from environment variable ROBOTMK_common_context."""

        self._set_context(contextname, log_level)
        self._context.load_config(
            DEFAULTS, ymlfile=ymlfile, varfile=varfile, default_cfg=default_cfg
        )

        # self.run_default = self._context.run_default
        # execute and output are the two main functions of each context:
        self.execute = self._context.execute
        self.output = self._context.output

    @property
    def config(self):
        return self._context.config

    def _set_context(self, contextname: str, log_level: str = None) -> None:
        """Sets the context of the Robotmk instance (=strategy)."""
        if contextname is None:
            contextname = os.environ.get("ROBOTMK_common_context", None)
        if contextname is None:
            raise ValueError(
                "No context given on CLI or set by environment variable ROBOTMK_common_context."
            )
        self._context = ContextFactory(contextname, log_level).get_context()
        self.config.set("common.context", contextname)

        # TODO: Setup Logging here


# print('Press Ctrl+{0} to exit'.format('Break' if os.name == 'nt' else 'C'))

# try:
#     # This is here to simulate application activity (which keeps the main thread alive).
#     while True:
#         time.sleep(2)
# except (KeyboardInterrupt, SystemExit):
#     # Not strictly necessary if daemonic mode is enabled but should be done if possible
#     scheduler.shutdown()



import json
from pathlib import Path
import base64
import zlib
from robotmk import __version__


class RFState:
    def __init__(self, target) -> None:
        self.target = target

    @property
    def statedata(self) -> dict:
        return {
            "suiteuname": self.target.suiteuname,
            "uuid": self.target.uuid,
            "rc": self.target.rc,
            "start_time": self._start_time.isoformat(),
            "end_time": self._end_time.isoformat(),
            "runtime": self._runtime.total_seconds(),
            "piggybackhost": self.target.config.get("suitecfg.piggybackhost", None),
            "output": {
                "html": {
                    "path": self.target.log_html_fullpath,
                },
                "xml": {
                    "path": self.target.output_xml_fullpath,
                    "base64": self.encode(
                        self.read_file(self.target.output_xml_fullpath)
                    ),
                },
                "console": self.target.console_results,
            },
            "robotmk_version": __version__,
        }

    def write(self) -> None:
        """Writes the console output to logdir and result JSON file to outputdir."""
        self.write_console_log()
        self.write_result_json()

    def timer_start(self) -> None:
        self._start_time = self.target._get_now_as_dt()

    def timer_stop(self) -> None:
        self._end_time = self.target._get_now_as_dt()
        self._runtime = self._end_time - self._start_time

    def write_console_log(self) -> None:
        for k, result in self.target.console_results.items():
            # HEREIWAS:
            data = json.dumps(result, indent=4)
            filename = (
                Path(self.target.outputdir)
                / f"{self.target.output_filename}-{int(k)}.txt"
            )
            self.write_data2file(data, filename)
            pass

    def write_result_json(self) -> None:
        """ """

        try:
            Path(self.target.statefile_fullpath).parent.mkdir(
                parents=True, exist_ok=True
            )
            with open(self.target.statefile_fullpath, "w", encoding="utf-8") as outfile:
                json.dump(self.statedata, outfile, indent=2, sort_keys=False)
        except IOError as e:
            # Error gets logged, will come to light by staleness check
            pass

    def write_data2file(self, data: str, filename: str):
        """Generic function to write data to a file."""
        try:
            with open(filename, "w") as f:
                f.write(data)
        except Exception as e:
            pass
            # TODO
            # raise RobotmkError("Failed to write data to file: %s" % e)

    def read_file(self, path, default=None):
        content = None
        try:
            with open(path, "r", encoding="utf-8") as file:
                content = file.read()
                if len(content) == 0:
                    # TODO: logging
                    # self.logwarn(
                    #     "File %s has no content, using defaults (%s)"
                    #     % (path, str(default))
                    # )
                    content = default
        except Exception as e:
            # TODO: logging
            # self.logwarn(
            #     "Error while reading %s (%s); using default (%s)"
            #     % (path, e, str(default))
            # )
            content = default
        return content

    def encode(self, data, encoding="zlib_codec"):
        # Caveat: to keep the zlib stream integrity, it must be converted to a
        # "safe" stream afterwards.
        # Reason: if there is a byte in the zlib stream which is a newline byte
        # by accident, Checkmk splits the byte string at this point - the
        # byte gets lost, stream integrity bungled.
        # Even if base64 blows up the data, this double encoding still saves space:
        # in:      692800 bytes  100    %
        # zlib:      4391 bytes    0,63 % -> compression 99,37%
        # base64:    5856 bytes    0,85 % -> compression 99,15%

        #    1. encode in UTF8
        #   2. compress with zlib
        #  3. encode with base64

        if encoding == "base64_codec":
            data_bytes = data.encode("utf-8")
            data_encoded = base64.b64encode(data_bytes)
            data_utf8 = data_encoded.decode("utf-8")
        elif encoding == "zlib_codec":
            data_bytes = data.encode("utf-8")
            data_zlib = zlib.compress(data_bytes, 9)
            data_encoded = base64.b64encode(data_zlib)
            data_utf8 = data_encoded.decode("utf-8")
        elif encoding == "utf_8":
            # nothing to do, already in utf8 = string
            data_utf8 = data
        else:
            # TODO: Catch the exception! (wrong encoding)!
            pass
        return data_utf8



from .abstract import AbstractExecutor


class Sequencer(AbstractExecutor):
    def __init__(self, config, *args, **kwargs):
        super().__init__(config)

    def run(self):
        """Start the sequencer, runs all API calls and ends"""
        print("Sequencer.run()")
        pass



"""CLI commands for execution of a single suite."""

import sys
import click
import json
from robotmk.cli.defaultgroup import DefaultGroup
from robotmk.main import Robotmk, DEFAULTS

# TODO: Refine the defaultgroup usage

#             _ _
#            (_) |
#   ___ _   _ _| |_ ___
#  / __| | | | | __/ _ \
#  \__ \ |_| | | ||  __/
#  |___/\__,_|_|\__\___|


# use module docstring as help text
@click.group(
    cls=DefaultGroup, default_if_no_args=True, help=__doc__, invoke_without_command=True
)
@click.pass_context
@click.option("--yml", "-y", help="Read config from custom YML file")
@click.option("--vars", "-v", help="Read vars from .env file (ignores environment)")
def suite(ctx, yml, vars):
    if vars and yml:
        raise click.BadParameter("Cannot use --yml and --vars at the same time")
    ctx_loglevel = ctx.parent.params.get("loglevel", DEFAULTS["common"]["log_level"])
    ctx.obj = Robotmk(
        contextname="suite", log_level=ctx_loglevel, ymlfile=yml, varfile=vars
    )


@suite.command(default=True)
@click.argument("suite", required=False)
@click.pass_context
def run(ctx, suite):
    """Trigger the start of a Robot Framework SUITE.

    SUITE must be a configuration subkey of the "suites" section.
    (can also be set by env:ROBOTMK_common_suiteuname.)
    """
    if suite:
        ctx.obj.config.set("common.suiteuname", suite)
    if bool(ctx.obj.config.get("common.suiteuname", None)):
        ctx.obj.execute()
    else:
        click.secho("Suite '%s' not found in configuration" % suite, fg="red")


@suite.command()
@click.argument("suite", required=False)
@click.pass_context
def result(ctx, suite):
    """Print the result JSON of a SUITE on STDOUT.

    SUITE must be a configuration subkey of the "suites" section.
    (can also be set by env:ROBOTMK_common_suiteuname.)

    To get the Checkmk Agent output, use "agent output" instead.
    """
    if suite:
        ctx.obj.config.set("common.suiteuname", suite)
    if bool(ctx.obj.config.get("common.suiteuname", None)):
        data = ctx.obj.output()
        click.secho(json.dumps(data, indent=4), fg="bright_white")

    else:
        click.secho("Suite '%s' not found in configuration" % suite, fg="red")


@suite.command()
@click.argument("suite", required=True)
@click.option(
    "--number",
    "-n",
    help="Number of last execution logs of SUITE to show",
    default=1,
    show_default=True,
)
@click.option(
    "--pid", "-p", help="Shows the execution log of SUITE with a specific PID"
)
@click.pass_context
def logs(ctx, suite, number, pid):
    """Display the log files of SUITE.

    SUITEID is eiher equal to the suite dir or a combination of suite dir and its unique tag as set in the configuration.
    Examples are: suite1, suite2_tagfoo, suite3_tagbaz
    (can also be set by env:ROBOTMK_common_suiteuname.)
    """
    click.secho("logs", fg="green")
    if int(number) != 1 and pid != None:
        raise click.BadParameter("Cannot use --number and --pid at the same time %d")
    click.echo("These are the logs of suite %s %d:" % (suite, number))
    pass


# TODO: implement this
@suite.command()
@click.argument("suite", required=False)
@click.pass_context
def shell(ctx, suite):
    """Open a shell in the RCC python environment of SUITE.

    SUITEID is eiher equal to the suite dir or a combination of suite dir and its unique tag as set in the configuration.
    Examples are: suite1, suite2_tagfoo, suite3_tagbaz
    (can also be set by env:ROBOTMK_common_suiteuname.)
    """
    click.secho("$> RCC shell of suite %s" % suite, fg="yellow")

    pass


@suite.command()
@click.pass_context
def vardump(ctx):
    click.secho("vardump", fg="green")
    # TODO: implement this
    pass


@suite.command(help="Dump the config as YML to STDOUT or FILE")
# add file arg
@click.argument("file", required=False, type=click.Path(exists=False))
@click.pass_context
def ymldump(ctx, file):
    click.secho(ctx.obj.config.to_yml(file), fg="bright_white")
    sys.exit(0)



from schema import Schema, SchemaError, And, Or, Use, Optional, Regex
import yaml


class RobotmkConfigSchema:
    """This class is used to validate the Robotmk config."""

    subdict_schema = Schema({str: str})
    failed_handling_schema = Schema({str: str})
    schema = Schema(
        {
            "common": {
                "context": Or("suite", "specialagent", "local"),
                # "suite": str,
                "cfgdir": str,
                "logdir": str,
                "tmpdir": str,
                "resultdir": str,
                "log_level": Or("debug", "info", "warning", "error", "critical"),
                "log_retention": int,
                Optional("suite"): str,
                Optional("k8s_auth"): {"user": str, "password": str, "url": str},
                Optional("robocorp_auth"): {"user": str, "password": str, "url": str},
            },
            "suites": {
                str: {
                    "path": str,
                    Optional("tag"): str,
                    Optional("piggybackhost"): str,
                    "run": Or(
                        {
                            "mode": "default",
                            "target": "local",
                            "rcc": Or(True, False),
                        },
                        {
                            "mode": Or(
                                "windows-1desktop", "windows-ndesktop", "linux-ndesktop"
                            ),
                            "target": "local",
                            "rcc": Or(True, False),
                            "user": str,
                            "password": str,
                        },
                    ),
                    # Optional("robot_params"): And(dict, Use(subdict_schema.validate)),
                    # Optional("failed_handling"): And(
                    #     dict, Use(failed_handling_schema.validate)
                    # ),
                    Optional("scheduling"): {"interval": int, "allow_overlap": bool},
                }
            },
        },
        ignore_extra_keys=True,
    )

    def __init__(self, data) -> None:
        self.data = data
        self.valid = None

    def validate(self) -> None:
        """Validate the config, set the valid property and return it."""
        try:
            self.valid = self.schema.validate(self.data)
        except SchemaError as e:
            self.valid = False
            self.error = e
        return self.valid



import yaml
from robotmk.config.config import Config
import os

cwd = os.path.dirname(__file__)
robotmk_yml = os.path.join(cwd, "robotmk.yml")
robotmk_env = os.path.join(cwd, "robotmk.env")


def test_suitecfg_shorthand():
    """Tests the 'suitecfg' shorthand which allows to set/get the config of the current
    suite by using the 'suitecfg' key."""
    cfg = Config()
    cfg.set("common.suiteuname", "foosuite")
    cfg.set("suitecfg.foo.bar", "baz")
    assert cfg.configdict["suites"]["foosuite"]["foo"]["bar"] == "baz"
    assert cfg.get("suitecfg.foo.bar") == "baz"


def test_path_prefix():
    """Tests if logdir, tmpdir, resultdir and robotdir are prefixed with the
    path given in the path_prefix key.
    This is not needed at all for Windows and not needed on Linux (as there is no common path).
    But for development, it is nice to have a common path prefix for all paths.
    """
    os.environ[
        "ROBOTMK_common_path__prefix"
    ] = "/home/simonmeggle/Documents/01_dev/rmkv2"
    os.environ["ROBOTMK_common_logdir"] = "agent/log/robotmk/logs"
    os.environ["ROBOTMK_common_tmpdir"] = "agent/tmp/robotmk"
    os.environ["ROBOTMK_common_resultdir"] = "agent/log/robotmk/results"
    os.environ["ROBOTMK_common_robotdir"] = "agent/robots"
    cfg = Config()
    # read variables from environment
    cfg.read_cfg_vars(path=None)
    assert (
        cfg.get("common.logdir")
        == "/home/simonmeggle/Documents/01_dev/rmkv2/agent/log/robotmk/logs"
    )
    assert (
        cfg.get("common.tmpdir")
        == "/home/simonmeggle/Documents/01_dev/rmkv2/agent/tmp/robotmk"
    )
    assert (
        cfg.get("common.resultdir")
        == "/home/simonmeggle/Documents/01_dev/rmkv2/agent/log/robotmk/results"
    )
    assert (
        cfg.get("common.robotdir")
        == "/home/simonmeggle/Documents/01_dev/rmkv2/agent/robots"
    )


def test_env2cfg_values():
    """Tests the conversion of environment variables to a config dictionary."""
    # simple values
    os.environ["ROBOTMK_suites_foo1"] = "qux"
    os.environ["ROBOTMK_common_log__level"] = "INFO"
    os.environ["ROBOTMK_common_suiteuname"] = "foo_suite"
    os.environ["ROBOTMK_suites_foo__suite_my__value"] = "qux"
    os.environ["ROBOTMK_suites_foo__suite_run_rcc"] = "false"
    cfg = Config()
    # read variables from environment
    cfg.read_cfg_vars(path=None)
    assert str(cfg.configdict["common"]["log_level"]) == "INFO"
    assert str(cfg.configdict["common"]["suiteuname"]) == "foo_suite"
    assert str(cfg.configdict["suites"]["foo_suite"]["my_value"]) == "qux"
    assert cfg.configdict["suites"]["foo_suite"]["run"]["rcc"] == False


def test_env2cfg_dicts():
    """Tests the conversion of environment variables to a config dictionary."""
    # # dicts
    os.environ["ROBOTMK_suites_foo__suite_my__dict_foo"] = "one"
    os.environ["ROBOTMK_suites_foo__suite_my__dict_bar"] = "two"
    os.environ["ROBOTMK_suites_foo__suite_my__dict_baz"] = "three"
    cfg = Config()
    # read variables from environment
    cfg.read_cfg_vars(path=None)

    assert cfg.configdict["suites"]["foo_suite"]["my_dict"] == {
        "foo": "one",
        "bar": "two",
        "baz": "three",
    }


def test_env2cfg_lists():
    """Tests the conversion of environment variables to a config dictionary."""
    # # # lists
    os.environ["ROBOTMK_suites_foo__suite_my__list_0"] = "one"
    os.environ["ROBOTMK_suites_foo__suite_my__list_1"] = "two"
    os.environ["ROBOTMK_suites_foo__suite_my__list_2"] = "three"
    cfg = Config()
    # read variables from environment
    cfg.read_cfg_vars(path=None)
    assert cfg.configdict["suites"]["foo_suite"]["my_list"] == [
        "one",
        "two",
        "three",
    ]


def test_env2cfg_listofdicts():
    """Tests the conversion of environment variables to a config dictionary."""
    # list of dicts
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_foo"] = "one"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_bar"] = "two"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_baz"] = "three"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_foo"] = "one"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_bar"] = "two"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_baz"] = "three"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_foo"] = "one"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_bar"] = "two"
    os.environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_baz"] = "three"
    cfg = Config()
    # read variables from environment
    cfg.read_cfg_vars(path=None)
    assert cfg.configdict["suites"]["foo_suite"]["my_list_of_dict"] == [
        {"foo": "one", "bar": "two", "baz": "three"},
        {"foo": "one", "bar": "two", "baz": "three"},
        {"foo": "one", "bar": "two", "baz": "three"},
    ]


def test_defaults():
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2}})
    assert cfg.configdict["common"]["a"] == 1
    assert cfg.configdict["common"]["b"] == 2


def test_read_yml_cfg():
    """Tests if parts of the default config are overwritten by the yml config.
    The default config sets a=1 and b=2, yml changes b to 3."""
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2}})
    cfg.read_yml_cfg(path=robotmk_yml)
    # unchanged
    assert cfg.configdict["common"]["a"] == 1
    # changed
    assert cfg.configdict["common"]["b"] == 3


def test_read_var_env_cfg():
    """Tests if parts of the default config are overwritten by
    - YML config
    - and then again overwritten by environment variables
    The default config sets a=1, b=2, c=3.
    Yml changes b to 3.
    Env changes c to 4."""
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2, "c": 3}})
    cfg.read_yml_cfg(path=robotmk_yml)
    # TODO: remove this line, it is only for testing
    os.environ["ROBOTMK_common_c"] = "4"
    os.environ["ROBOTMK_common_cc"] = "4"
    os.environ["ROBOTMK_common_cff"] = "4"
    os.environ["ROBOTMK_common_chhh"] = "4"
    # read variables from environment
    cfg.read_cfg_vars(path=None)
    # unchanged
    assert str(cfg.configdict["common"]["a"]) == "1"
    assert str(cfg.configdict["common"]["b"]) == "3"
    # changed
    assert str(cfg.configdict["common"]["c"]) == "4"


def test_read_var_file_cfg():
    """Tests if parts of the default config are overwritten by
    - YML config
    - variables in a file
    - and then again overwritten by env vars
    The default config sets a=1, b=2, c=3.
    Yml changes b to 3.
    File changes c to 5.
    Env changes c to 55."""
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2, "c": 3}})
    cfg.read_yml_cfg(path=robotmk_yml)
    os.environ["ROBOTMK_common_c"] = "55"
    cfg.read_cfg_vars(path=robotmk_env)
    # unchanged
    assert str(cfg.configdict["common"]["a"]) == "1"
    assert str(cfg.configdict["common"]["b"]) == "3"
    # changed
    assert str(cfg.configdict["common"]["c"]) == "55"


def test_read_var_file_added_cfg():
    """Tests if parts of the default config are overwritten by
    - YML config
    - variables in a file
    - env vars
    - and then again overwritten by added vars
    The default config sets a=1, b=2, c=3.
    Yml changes b to 3.
    File changes c to 5.
    Env changes c to 55.
    Added var changes c to 6."""
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2, "c": 3}})
    cfg.read_yml_cfg(path=robotmk_yml)
    os.environ["ROBOTMK_common_c"] = "55"
    cfg.read_cfg_vars(path=robotmk_env)
    cfg.set("common.c", 66)
    cfg.set("common.d.e.f.g", 77)
    # unchanged
    assert str(cfg.configdict["common"]["a"]) == "1"
    assert str(cfg.configdict["common"]["b"]) == "3"
    # changed
    assert str(cfg.configdict["common"]["c"]) == "66"
    assert str(cfg.configdict["common"]["d"]["e"]["f"]["g"]) == "77"


def test_config_to_yml():
    """Tests if the config can be dumped to a valid YML string"""
    cfg = Config()
    cfg.set_defaults({"common": {"a": 1, "b": 2, "c": 3}})
    cfg.read_yml_cfg(path=robotmk_yml)
    os.environ["ROBOTMK_common_c"] = "4"
    os.environ["ROBOTMK_foo_bar_x"] = "44"
    cfg.read_cfg_vars(path=None)
    # read variables from a file
    cfg.read_cfg_vars(path=robotmk_env)
    # unchanged
    assert str(cfg.configdict["common"]["a"]) == "1"
    assert str(cfg.configdict["common"]["b"]) == "3"
    # changed
    assert str(cfg.configdict["common"]["c"]) == "4"
    yml_str = cfg.to_yml()
    yml = yaml.load(yml_str, Loader=yaml.Loader)
    assert str(yml["foo"]["bar"]["x"]) == "44"


def test_envvar2dict():
    """Tests if only env vars with the prefix ROBOTMK_ are read into the config"""
    cfg = Config()
    os.environ["ROBOTMK_foo_bar1"] = "1"
    os.environ["ROBOTMK_foo_bar2"] = "2"
    os.environ["ROBOTMK_foo_bar3"] = "3"
    os.environ["DONT_foo_bar4"] = "4"
    cfg.read_cfg_vars(path=None)
    assert str(cfg.configdict["foo"]["bar1"]) == "1"
    assert str(cfg.configdict["foo"]["bar2"]) == "2"
    assert str(cfg.configdict["foo"]["bar3"]) == "3"
    # assert not
    assert not "bar4" in cfg.configdict["foo"]


def test_dotcfg2env():
    """Tests if the setenv() function works"""
    environ = {}
    dotstrings = {
        "common.logdir": "/another/path",
        "suitecfg.uuid": "1234",
        "suitecfg.run.rcc": False,
    }
    cfg = Config()
    cfg.set("common.suiteuname", "foo_suite")
    cfg.dotcfg_to_env(dotstrings, environ=environ)
    assert environ["ROBOTMK_common_logdir"] == "/another/path"
    assert environ["ROBOTMK_suites_foo__suite_uuid"] == "1234"
    assert environ["ROBOTMK_suites_foo__suite_run_rcc"] == "false"


def test_cfg2env():
    """Tests the conversion of a config dictionary to environment variables."""
    environ = {}
    cfg = Config()
    cfg.set("common.log_level", "INFO")
    cfg.set("common.suiteuname", "foo_suite")
    cfg.set("suitecfg.my_value", "qux")
    cfg.set("suitecfg.run.rcc", False)
    cfg.set("suitecfg.my_list", ["one", "two", "three"])
    cfg.set("suitecfg.my_dict", {"foo": "one", "bar": "two", "baz": "three"})
    cfg.set(
        "suitecfg.my_list_of_dict",
        [
            {"foo": "one", "bar": "two", "baz": "three"},
            {"foo": "one", "bar": "two", "baz": "three"},
            {"foo": "one", "bar": "two", "baz": "three"},
        ],
    )
    cfg.cfg_to_environment(cfg.configdict, environ=environ)
    # simple values
    assert environ["ROBOTMK_common_log__level"] == "INFO"
    assert environ["ROBOTMK_common_suiteuname"] == "foo_suite"
    assert environ["ROBOTMK_suites_foo__suite_my__value"] == "qux"
    assert environ["ROBOTMK_suites_foo__suite_run_rcc"] == "false"
    # lists
    assert environ["ROBOTMK_suites_foo__suite_my__list_0"] == "one"
    assert environ["ROBOTMK_suites_foo__suite_my__list_1"] == "two"
    assert environ["ROBOTMK_suites_foo__suite_my__list_2"] == "three"
    # dicts
    assert environ["ROBOTMK_suites_foo__suite_my__dict_foo"] == "one"
    assert environ["ROBOTMK_suites_foo__suite_my__dict_bar"] == "two"
    assert environ["ROBOTMK_suites_foo__suite_my__dict_baz"] == "three"
    # list of dicts
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_foo"] == "one"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_bar"] == "two"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_0_baz"] == "three"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_foo"] == "one"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_bar"] == "two"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_1_baz"] == "three"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_foo"] == "one"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_bar"] == "two"
    assert environ["ROBOTMK_suites_foo__suite_my__list__of__dict_2_baz"] == "three"


# TODO: split_varstring
# TODO: config validation



from .abstract import Target
from ..strategies import RunStrategy
from robotmk.logger import RobotmkLogger


class RemoteTarget(Target):
    def __init__(self, suiteuname: str, config: dict, logger: RobotmkLogger):
        super().__init__(suiteuname, config, logger)

    def run(self):
        print("NOT YET IMPLEMENTED")
        pass

    def output(self):
        # return a dummy dict from a dummy REST API call
        # TODO: implement REST API Call to Robocorp to fetch the output
        return {
            "suiteuname": self.suiteuname,
            "uuid": "5227a230696f40939a51d60dc1cbdd29",
            "rc": 1,
            "start_time": "2023-04-25T16:58:34.573492+02:00",
            "end_time": "2023-04-25T16:58:56.965265+02:00",
            "runtime": 22.391773,
            "piggybackhost": self.piggybackhost,
            "output": {
                "html": {
                    "path": "/home/simonmeggle/Documents/01_dev/rmkv2/agent/log/robotmk/robotframework/rf_suite_default_1682434706_2aa6fe6d.html"
                },
                "xml": {
                    "path": "/home/simonmeggle/Documents/01_dev/rmkv2/agent/log/robotmk/robotframework/rf_suite_default_1682434706_2aa6fe6d.xml",
                    "base64": "eNrlVU1PGzEQvfMrLB9QWynrbIBAIVmUfqAiAUWQnitnd7qx4o+V7U3g39fj3YSEBFSp5dST7fHMe/PGY3tw/qAkmYN1wughTZMuJaBzUwhdDumP8UXnhJ5newNrJsaTEjRY7o0d0jtAQz9JeUre3T76qdHkIEm7SZ+EmRS6fnhPlwFQDGmv2zvoHvaOSNo/PTo5PeonHw+PKbEVH1Jva6DE5VNQfJXLIQ28rhYeiAjxLqVEcwVhhrZR8De1zcOaTY0C5oQyWkFZSmBfTF4r0N6xbvqzgDmzajbvMR7S8SxKcaxFecbReaIZczdz/4aFecRKog0ZPTj/RNjxK85roYXikqA/xSoGWx8jZovW48qUZGzIZ6OdkYA+E8vt45B+qoX0lxqduS2zu3hgqIEsuCPwAHkdziEZMNzdGxQmzwKWI34KpBRz0ESBcyF54k005g1FiEDfUCbPfe1IMwzp7ej+nuLKei8wsx0HnPaxm4rX9llAZrPFusQLLuRuYZgI7jZZxyouhJ/u0MB1QUzlQyNxKR8Jlz60FRE+RPLSrTQpVxLMLshQ1e4UQ49KmIMMeY0ur2g2cg4sAn+11tgBCxDbxYmuf1uc47XivA1+9uHb+PrqA9mX/sxVXJNcchfgFdgS6H7pz2IDTUMDTSBU10Jn2UixwhZcLUNNo3+RIAxDHIzExdQuZ+voGhadRkikuIFFq+t0A2Arq1/YF7iD6re4Ji9ytT2xImvXz9g2z3UTcbcOI4t1Hd9l8XY6kGtdB5L9nzoGrEkOrwa+AO2zMAZVyfDTkOanUlxoEh/fl96vrSt0w0ZrF6ZZSV45aC1pcx0j6NvACedFjsK88Vy2RlLFwoV/GeuNccTNRIWWbCRlfOJdU5RYkjY0PHNxGYd1klfwdv+zWTOuKP4MZOsjbWGSZ/kuU2MbBQA883jET7N4sNneb5Vz1Zw=",
                },
            },
        }



from pathlib import Path
from robotmk.logger import RobotmkLogger
from .abstract import Target

from .local.robotframework import RobotFrameworkTarget
from .local.rcc import RCCTarget
from .remote import RemoteTarget


class TargetFactory:
    def __init__(self, suiteuname: str, config, logger: RobotmkLogger):
        self.suiteuname = suiteuname
        self.config = config
        self.logger = logger

    def create(self) -> Target:
        """Create a target object."""
        otarget = None
        suitecfg = self.config.get("suites.%s" % self.suiteuname)
        if suitecfg == None:
            self.error("Suite '%s' is not part of the config!" % self.suiteuname)

        target_name = suitecfg.get("run.target")
        if target_name == "local":
            path = Path(self.config.get("common.robotdir")).joinpath(
                suitecfg.get("path")
            )
            if path.exists():
                # Watch the correct boolean interpretation especially for the
                # case of "run.rcc" being set to "False" in the config file.
                # Danger of starting RCC inside RCC inside RCC!
                if suitecfg.get("run.rcc", False):
                    otarget = RCCTarget(self.suiteuname, self.config, self.logger)
                else:
                    otarget = RobotFrameworkTarget(
                        self.suiteuname, self.config, self.logger
                    )
            else:
                # TBD: check this if this gets logged...
                self.error("Suite path does not exist: " + str(path))
        elif target_name == "remote":
            otarget = RemoteTarget(self.suiteuname, self.config, self.logger)
        else:
            self.error("Unknown target type for suite %s!" % self.suiteuname)
        return otarget



from abc import ABC, abstractmethod
from pathlib import Path
import json
from ..strategies import RunStrategy, RunStrategyFactory
from uuid import uuid4

from robotmk.logger import RobotmkLogger


class Target(ABC):
    """A Target defines the environment where a suite gets executed.

    It's the abstraction of either
    - a local Robot suite or ("target: local")
    - an API call to an external platform ("target: remote") like Robocorp or Kubernetes
    """

    def __init__(self, suiteuname: str, config, logger: RobotmkLogger):
        self.suiteuname = suiteuname
        self.config = config

        self.commoncfg = self.config.get("common")

        self._logger = logger
        if not self._logger is None:
            self.debug = self._logger.debug
            self.info = self._logger.info
            self.warning = self._logger.warning
            self.error = self._logger.error
            self.critical = self._logger.critical

    @property
    def piggybackhost(self):
        return self.config.get("suitecfg.piggybackhost", "localhost")

    @abstractmethod
    def run(self):
        """Abstract method to run a suite/target."""
        pass

    @abstractmethod
    def output(self):
        """Abstract method to get the output of a suite/target."""
        pass



"""CLI commands for the specialagent context."""
import sys
import click
from robotmk.main import Robotmk, DEFAULTS


# use module docstring as help text
@click.group(help=__doc__, invoke_without_command=True)
@click.pass_context
@click.option("--vars", "-v", help="Read vars from .env file (ignores environment)")
def specialagent(ctx, vars):
    click.echo("Executing specialagent....")
    ctx.obj = Robotmk("specialagent", vars=vars)
    ctx.obj.config.set("common.context", "specialagent")

    pass


@specialagent.command()
@click.pass_context
def sequencer(ctx):
    """Start the sequencer to execute tests once, respecting their interval."""
    ctx.obj.execute()


@specialagent.command()
@click.pass_context
def output(ctx):
    """Produces Checkmk Agent output for all suite results."""
    ctx.obj.output()
    pass


@specialagent.command(help="Dump the config as YML to STDOUT or FILE")
# add file arg
@click.argument("file", required=False, type=click.Path(exists=False))
@click.pass_context
def ymldump(ctx, file):
    click.secho(ctx.obj.config.to_yml(file), fg="bright_white")
    sys.exit(0)



from abc import ABC, abstractmethod
from robotmk.config import Config
from robotmk.logger import RobotmkLogger
from pathlib import Path


class AbstractContext(ABC):
    """Abstract class for context objects. Context objects are used to
    encapsulate the different contexts in which robotmk can be run (local, specialagent, suite).
    """

    def __init__(self):
        # self._config_factory = Config()
        self._logger = None
        self.init_logger()
        self.config = Config()

    @property
    def logger(self):
        # self.__init_logger()
        return self._logger

    def init_logger(self):
        # initialize the logger only when config was loaded
        if self._logger is None and getattr(self, "config", None):
            self._logger = RobotmkLogger(
                Path(self.config.get("common.logdir")).joinpath("robotmk.log"),
                self.config.get("common.log_level"),
            )
            self.debug = self._logger.debug
            self.info = self._logger.info
            self.warning = self._logger.warning
            self.error = self._logger.error
            self.critical = self._logger.critical

    @abstractmethod
    def load_config(self, defaults, **kwargs) -> None:
        """Depening on the context strategy, the config object loads cfg from different sources."""
        raise NotImplementedError("Subclass must implement abstract method")

    @abstractmethod
    def refresh_config(self):
        """Load the config again, e.g. after a change in the config file."""
        raise NotImplementedError("Subclass must implement abstract method")

    # @abstractmethod
    # def run_default(self):
    #     """Encapsulates everything that needs to be done to
    #     run robotmk when it is run only with context, but without subcommand."""
    #     raise NotImplementedError("Subclass must implement abstract method")

    @abstractmethod
    def execute(self, *args, **kwargs):
        """Encapsulates everything that needs to be done to run robotmk."""
        raise NotImplementedError("Subclass must implement abstract method")

    @abstractmethod
    def output(self):
        """Encapsulates everything that needs to be done to produce agent output."""
        raise NotImplementedError("Subclass must implement abstract method")



"""This module encapsulates everyhting related so a single suite 
execution, either against a local target (RCC/RF) or against a remote one (API)."""

from uuid import uuid4
from pathlib import Path
from ..abstract import AbstractContext

from robotmk.config import RobotmkConfigSchema
from .target.abstract import Target
from .target.target_factory import TargetFactory


class SuiteContext(AbstractContext):
    def __init__(self):
        super().__init__()
        # the target object
        self._otarget = None
        self._ymlschema = RobotmkConfigSchema

    @property
    def suiteuname(self):
        """suiteuname under "common" sets the suite to start (suitename + tag)"""
        try:
            suiteuname = self.config.get("common.suiteuname")
        except AttributeError:
            pass
            # TODO: What if suite is not found?
        return suiteuname

    @property
    def target(self) -> Target:
        # singleton
        return self._otarget

    def get_target(self) -> Target:
        if not self._otarget:
            self.init_logger()
            self._otarget = TargetFactory(
                self.suiteuname, self.config, self.logger
            ).create()
        return self._otarget

    def load_config(self, defaults, **kwargs) -> None:
        """Load the config for suite context.

        Suite context can merge the config from
        - OS defaults
        - + environment variables
        - + YML file (default/custom = --yml)
        - + var file (= --vars)
        - + environment variables


        """
        if kwargs.get("default_cfg", {}):
            # Suite was started by scheduler, config was passed
            self.config.configdict = kwargs["default_cfg"]
        else:
            # After the defaults are read...
            self.config.set_defaults(defaults)
            # ...lets first search for env vars which point to another location of the robotmk.yml!
            self.config.read_cfg_vars(path=None)
            self.config.read_yml_cfg(path=kwargs["ymlfile"], must_exist=False)
            # In the end, environment variables can override everything
            self.config.read_cfg_vars(path=kwargs["varfile"])

        # TODO: validate later so that config can be dumped
        # self.config.validate(self._ymlschema)

    def refresh_config(self) -> bool:
        """Re-loads the config and returns True if it changed"""
        # TODO: implement this
        pass

    # def run_default(self):
    #     """Implements the default action for suite context."""
    #     # TODO: execute one single suite
    #     print("Suite context default action = execute single suite ")
    #     pass

    def execute(self):
        """Runs a single suite, either fs/remote."""
        # TODO: is it better to pass the suitename to get_target()?
        self.get_target().run()

    def output(self):
        """Gathers the result of the given suite (fs/remote) and returns it"""
        return self.get_target().output()



"""CLI commands for the agent context. 

Executes the Robotmk scheduler on Windows & Linux, produces Checkmk Agent output"""
import sys
import click
from robotmk.main import Robotmk, DEFAULTS


#                         | |
#    __ _  __ _  ___ _ __ | |_
#   / _` |/ _` |/ _ \ '_ \| __|
#  | (_| | (_| |  __/ | | | |_
#   \__,_|\__, |\___|_| |_|\__|
#          __/ |
#         |___/


# use module docstring as help text
@click.group(help=__doc__, invoke_without_command=True)
@click.pass_context
@click.option("--yml", "-y", help="Read config from custom YML file")

# @click.option("--vars", "-v", help="Read vars from .env file (ignores environment)")
def agent(ctx, yml):
    ctx_loglevel = ctx.parent.params.get("loglevel", DEFAULTS["common"]["log_level"])
    ctx.obj = Robotmk(contextname="agent", log_level=ctx_loglevel, ymlfile=yml)
    if ctx.invoked_subcommand is None:
        click.secho("No subcommand given. Use --help for help.", fg="red")
        sys.exit(1)


@agent.command()
@click.pass_context
def scheduler(ctx):
    """Starts the scheduler to execute tests repeatedly."""
    ctx.obj.execute()


@agent.command()
@click.pass_context
def output(ctx):
    """Emits Checkmk Agent output for all suite results."""
    data = ctx.obj.output()
    click.secho(data, fg="bright_white")


@agent.command(help="Dumps the config as YML to STDOUT or FILE")
# add file arg
@click.argument("file", required=False, type=click.Path(exists=False))
@click.pass_context
def ymldump(ctx, file):
    click.secho(ctx.obj.config.to_yml(file), fg="bright_white")
    sys.exit(0)



import re
import os
from click.testing import CliRunner
import robotmk.cli.cli as cli

cwd = os.path.dirname(__file__)
robotmk_yml = os.path.join(cwd, "robotmk.yml")
robotmk_env = os.path.join(cwd, "robotmk.env")


# test help message
def test_specialagent_cli_help():
    """The help message should contain the three contexts and the help message itself."""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["specialagent", "--help"])
    assert result.exit_code == 0
    # assert "Robotmk CLI Interface." in result.output
    assert not re.search(r"--yml TEXT", result.output, re.DOTALL)
    assert re.search(r"--vars TEXT", result.output, re.DOTALL)
    assert re.search(r"Commands:.*output.*sequencer", result.output, re.DOTALL)



from abc import ABC, abstractmethod


class AbstractExecutor:
    """Base class for the executor.

    This is the abstraction class for the executors of Scheduler/Sequencer"""

    def __init__(self, config, *args, **kwargs):
        self.config = config

    @abstractmethod
    def execute(self):
        """Abstract method for the default run action of the executor"""
        pass



from abc import ABC, abstractmethod
from pathlib import Path
from uuid import uuid4
import json

from ..abstract import Target
from ...strategies import RunStrategyFactory
from robotmk.logger import RobotmkLogger


class LocalTarget(Target):
    """A FS target is a single RF suite or a RCC task, ready to run from the local filesystem.

    It also encapsulates the implementation details of the RUN strategy, which is
    either a headless or a headed execution (RDP, XVFB, Scheduled Task)."""

    def __init__(
        self,
        suiteuname: str,
        config: dict,
        logger: RobotmkLogger,
    ):
        super().__init__(suiteuname, config, logger)
        self.path = Path(self.config.get("common.robotdir")).joinpath(
            self.config.get("suitecfg.path")
        )
        # TODO: run strategy should not be set in init, because output() always reads results from filesystem
        self.run_strategy = RunStrategyFactory(self).create()
        # list of subprocess' results and console output
        self.console_results = {}

    @abstractmethod
    def run(self):
        """Implementation in subclasses RCCTarget and RobotFrameworkTarget"""
        pass

    def output(self):
        """Read the result artifacts from the filesystem."""
        try:
            with open(self.statefile_fullpath) as f:
                data = json.load(f)
                return data
        except FileNotFoundError:
            # return an empty dict to indicate that the file is not present
            return {}

    @property
    def pre_command(self):
        return None

    @property
    def main_command(self):
        return None

    @property
    def post_command(self):
        return None

    @property
    def uuid(self):
        """If a UUID is already part of the suite config, use this. Otherwise generate a new one.

        The idea is that the UUID is handed over between all robotmk calls and lastly part of the
        result JSON."""
        uuid_ = self.config.get("suitecfg.uuid", False)
        if not uuid_:
            uuid_ = uuid4().hex
        return uuid_

    @property
    def logdir(self):
        return self.config.get("common.logdir")

    @property
    def resultdir(self):
        return Path(self.logdir).joinpath("results")

    @property
    def statefile_fullpath(self):
        return str(Path(self.resultdir).joinpath(self.suiteuname + ".json"))

    @property
    def is_disabled_by_flagfile(self):
        """The presence of a file DISABLED inside of a Robot suite will prevent
        Robotmk to execute the suite, either by RCC or RobotFramework."""
        return self.path.joinpath("DISABLED").exists()



from abc import ABC, abstractmethod
import platform
from robotmk.logger import RobotmkLogger
import subprocess
import os
from dataclasses import dataclass, field, asdict
from typing import List

# TODO: split this into modules
# TODO:


@dataclass
class Result:
    """Result of a subprocess execution."""

    args: List[str] = field(default_factory=list)
    returncode: int = 0
    stdout: List[str] = field(default_factory=list)
    stderr: List[str] = field(default_factory=list)


class RunStrategy(ABC):
    def __init__(self, target) -> None:
        self.target = target

        # self.suiteuname = suiteuname
        # self.config = config
        # self._logger = logger
        # self.debug = self._logger.debug
        # self.info = self._logger.info
        # self.warning = self._logger.warning
        # self.error = self._logger.error
        # self.critical = self._logger.critical

    def run(self, *args, **kwargs):
        """Template method which bundles the linked methods to run.

        The concrete strategy selectivly overrides the methods to implement."""
        rc = max(
            self.exec_pre(*args, **kwargs),
            self.exec_main(*args, **kwargs),
            self.exec_post(*args, **kwargs),
        )
        return rc

    @abstractmethod
    def exec_pre(self, *args, **kwargs) -> int:
        """Prepares the given suite."""
        pass

    @abstractmethod
    def exec_main(self, *args, **kwargs) -> int:
        """Execute the the given suite."""
        pass

    @abstractmethod
    def exec_post(self, *args, **kwargs) -> int:
        """Cleans up the given suite."""
        pass


class Runner(RunStrategy):
    """This Strategy is the only one which executes a 'job' in fact.

    - run a Robot Framework Suite
    - run a RCC task
    """

    def __init__(self, target) -> None:
        super().__init__(target)

    def run_subprocess(self, command, environ) -> Result:
        """If command was given, run the subprocess and return the result object."""
        if command:
            res = subprocess.run(command, capture_output=True, env=environ)
            return Result(
                args=res.args,
                returncode=res.returncode,
                stdout=res.stdout.decode("utf-8").splitlines(),
                stderr=res.stderr.decode("utf-8").splitlines(),
            )
        else:
            return Result()

    def exec_pre(self, *args, **kwargs) -> int:
        result = self.run_subprocess(self.target.pre_command, os.environ)
        return result.returncode

    def exec_main(self, *args, **kwargs) -> int:
        # DEBUG: " ".join(self.target.main_command)

        # DEBUG: [f"{k}={v}" for (k,v) in environment.items()  if k.startswith("RO")]
        # DEBUG: "; ".join([f"export {k}={v}" for (k,v) in kwargs.get("env").items() if k.startswith("RO")])
        result = self.run_subprocess(
            self.target.main_command, kwargs.get("env", os.environ)
        )

        # TODO: log console output? Save it anyway because a a fatal RF error must be tracable.
        # RCC does not re.execute...
        if getattr(self.target, "attempt", None) is None:
            self.target.attempt = 1
        self.target.console_results[self.target.attempt] = asdict(result)
        return result.returncode

    def exec_post(self, *args, **kwargs) -> int:
        result = self.run_subprocess(self.target.post_command, os.environ)
        return result.returncode


class WindowsTask(RunStrategy):
    """Parent class for Single and Multi desktop strategies.

    Both have in common that they need to create a scheduled task."""

    def __init__(self, target) -> None:
        super().__init__(target)

    @abstractmethod
    def exec_pre(self, *args, **kwargs) -> int:
        pass

    @abstractmethod
    def exec_main(self, *args, **kwargs) -> int:
        pass

    @abstractmethod
    def exec_post(self, *args, **kwargs) -> int:
        pass


class WindowsSingleDesktop(WindowsTask):
    """Concrete class to run a suite with UI on Windows.

    ....
    """

    def __init__(self, target) -> None:
        super().__init__(target)

    def exec_pre(self, *args, **kwargs) -> int:
        # create the scheduled task for the given user
        pass

    def exec_main(self, *args, **kwargs) -> int:
        # run schtask.exe to run the task
        pass

    def exec_post(self, *args, **kwargs) -> int:
        pass


class WindowsMultiDesktop(WindowsTask):
    """Concrete class to run a suite in a loopback RDP session.

    This will require a Windows Server with RDP enabled and a proper
    MSTC license. Although there is https://github.com/stascorp/rdpwrap
    (https://www.anyviewer.com/how-to/windows-10-pro-remote-desktop-multiple-users-0427.html)
    """

    def __init__(self, target) -> None:
        super().__init__(target)

    def exec_pre(self, *args, **kwargs) -> int:
        # create RDP file:
        # rdp_file = "loopback.rdp"
        # with open(rdp_file, "w") as f:
        #     f.write(f"""\
        # username:s:{username}
        # password 51:b:{password}
        # full address:s:127.0.0.2
        # """)
        pass

    def exec_main(self, *args, **kwargs) -> int:
        # Launch the RDP session with the specified command
        # os.system(f"mstsc /v:127.0.0.2 /f /w:800 /h:600 /v:127.0.0.2 /u:{username} /p:{password} /v:{rdp_file} /start:{command}")
        # os.system(f'mstsc /v:127.0.0.1 /f /w:800 /h:600 /u:{username} /p:{password} /v:127.0.0.1 /w:800 /h:600 /v:127.0.0.1 /w:800 /h:600 /admin /restrictedAdmin cmd /c "{command}"')

        pass

    def exec_post(self, *args, **kwargs) -> int:
        # Close the RDP session
        # os.system(f'tscon /dest:console')
        pass


class LinuxMultiDesktop(RunStrategy):
    """Executes a suite with a user interface on Linux."""

    def __init__(self, target) -> None:
        super().__init__(target)

    def exec_pre(self, *args, **kwargs) -> int:
        pass

    def exec_main(self, *args, **kwargs) -> int:
        pass

    def exec_post(self, *args, **kwargs) -> int:
        pass


#    __           _
#   / _|         | |
#  | |_ __ _  ___| |_ ___  _ __ _   _
#  |  _/ _` |/ __| __/ _ \| '__| | | |
#  | || (_| | (__| || (_) | |  | |_| |
#  |_| \__,_|\___|\__\___/|_|   \__, |
#                                __/ |
#                               |___/


class RunStrategyFactory:
    """Factory for creating the proper run strategy for a given suite/OS."""

    def __init__(self, target):
        self.target = target

    def create(self) -> RunStrategy:
        """Creates a run strategy based on the given parameters.

        Returns:
            RunStrategy: The run strategy to use.
        """
        mode = self.target.config.get(
            "suites.%s.run.mode" % self.target.config.get("common.suiteuname")
        )
        _platform = platform.system().lower()
        if mode == "default":
            return Runner(self.target)
        elif mode == "windows-1desktop" and _platform == "windows":
            return WindowsSingleDesktop(self.target)
        elif mode == "windows-ndesktop" and _platform == "windows":
            return WindowsMultiDesktop(self.target)
        elif mode == "linux-ndesktop" and _platform == "linux":
            return LinuxMultiDesktop(self.target)
        else:
            raise ValueError(
                "Invalid combination of platform (%s) and run mode (%s)."
                % (_platform, mode)
            )



from pathlib import Path

from ..abstract import AbstractContext

from robotmk.config.yml import RobotmkConfigSchema
from robotmk.executor.sequencer import Sequencer


class SpecialAgentContext(AbstractContext):
    def __init__(self):
        super().__init__()
        self.ymlschema = RobotmkConfigSchema

    @property
    def logger(self):
        if not self._logger:
            self._logger = RobotmkLogger(
                Path(self.config.common.logdir).joinpath("robotmk.log"),
                self.config.common.log_level,
            )
        return self._logger

    def load_config(self, defaults, **kwargs) -> None:
        """Load the config for specialagent context.

        This context can merge the config from
        - OS defaults
        - + var file (= --vars)
        - + environment variables

        (There is no YML file for specialagent context!)
        """
        self.config.set_defaults(defaults)
        self.config.read_cfg_vars(path=kwargs.get("path"))
        # TODO: validate later so that config can be dumped
        # self.config.validate(self.ymlschema)

    def refresh_config(self) -> bool:
        """Re-loads the config and returns True if it changed"""
        # TODO: implement this

    # def run_default(self):
    #     """Implements the default action for specialagent context."""
    #     # TODO: start the sequencer
    #     print("Specialagent context default action = trigger APIs and output")
    #     pass

    def execute(self, *args, **kwargs):
        """Implements the run action for specialagent context."""
        print("Specialagent context run action")
        self.executor = Sequencer(self.config)
        self.executor.run()

    def output(self):
        """Implements the agent output for local context."""
        print("Specialagent context output")
        self.outputter = SpecialAgentOutput(self.config)
        pass



from click.testing import CliRunner
import robotmk.cli.cli as cli
import re


# test help message
def test_cli_help():
    """The help message shoudl contain the three contexts and the help message itself."""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["--help"])
    assert result.exit_code == 0
    assert "Robotmk CLI Interface." in result.output
    assert re.search(r"Commands:.*agent.*specialagent.*suite", result.output, re.DOTALL)


def test_cli_invalid_yml_option():
    """No YML config file possible without a context."""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["--yml", "invalid_file"])
    assert result.exit_code == 2
    assert "No such option: --yml" in result.output


def test_cli_invalid_vars_option():
    """No vars file possible without a context."""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["--vars", "invalid_file"])
    assert result.exit_code == 2
    assert "No such option: --vars" in result.output



import os
from abc import ABC, abstractmethod
from robot.rebot import rebot
from pathlib import Path
import glob


class RetryStrategyFactory:
    """Factory for execution strategies"""

    def __init__(self, target):
        self.target = target
        self.run = self.target.run_strategy.run

    def create(self):
        """Create the execution strategy"""
        strategy = self.target.config.get("suitecfg.retry_failed.strategy", "complete")
        if strategy == "complete":
            return CompleteRetry(self.target)
        elif strategy == "incremental":
            return IncrementalRetry(self.target)
        else:
            raise Exception("Unknown retry strategy: %s" % strategy)


class RetryStrategy(ABC):
    """Execution strategy interface for suites"""

    def __init__(self, target):
        self.target = target
        self.target.attempt = 1

    @property
    def max_attempts(self):
        """Maximum number of attempts to execute a suite (1st + retries)"""
        return 1 + self.target.config.get("suitecfg.retry_failed.retry_attempts", 0)

    def run(self):
        """Run the suite and retry failed tests if necessary."""

        # TODO: set TIC

        for attempt in range(1, self.max_attempts + 1):
            # if self.max_attempts > 1:
            #     self._runner.loginfo(
            #         f" > Starting attempt {attempt}/max {max_exec} ({str(self.})"
            #     )
            # else:
            #     self._runner.loginfo(f" > Starting suite...")
            self.target.attempt = attempt

            # TODO: log the cli args
            rc = self.target.run_strategy.run()
            # TODO: Logging
            # if rc > 250:
            #     self.logerror(
            #         "RC > 250 = Robot exited with fatal error. There are no logs written."
            #     )
            if self.max_attempts == 1 or (self.target.attempt == 1 and rc == 0):
                # if only one attempt allowed or 1st attempt was OK, we are done
                break
            else:
                # more attempts allowed and 1st attempt was not OK
                if rc == 0:
                    # this retry was OK, get out here
                    self._finalize_results()
                    break
                else:
                    if self.target.attempt < self.max_attempts:
                        self._reparametrize()
                    else:
                        # ...GAME OVER! => MERGE
                        # TODO: logging
                        # self._runner.loginfo(
                        #     "   Even the last attempt was unsuccessful!"
                        # )
                        self._finalize_results()
        return rc

    @abstractmethod
    def _reparametrize(self):
        """Reparametrize the suite for the next attempt.

        Only incremental strategy needs to do something here."""
        pass

    def _finalize_results(self):
        """Merge the XML result files into a new final result"""

        # Attempt "None" sets output filenames without a attempt number
        self.target.attempt = None

        outputfiles = self._glob_target_outputfiles()
        filenames = [Path(f).name for f in outputfiles]
        # TODO: log the files to merge
        # TODO: use string fields of the subclasses for individual logging
        # for f in filenames:
        #     self.suite._runner.logdebug(" - %s" % f)

        # rebot wants to print out the generated file names on stdout; write to devnull
        devnull = open(os.devnull, "w")
        rebot(
            *outputfiles,
            outputdir=self.target.outputdir,
            output=self.target.output_xml,
            log=self.target.log_html,
            report=None,
            merge=True,
            stdout=devnull,
        )
        # self.suite._runner.loginfo("Merged results of all reexecutions into:")
        # self.suite._runner.loginfo(" - %s" % self.suite.output)
        # self.suite._runner.loginfo(" - %s" % self.suite.log)

    def _glob_target_outputfiles(self):
        """Returns the list of XML output files of the target execution attempts 1..n"""
        glob_pattern = "%s-*.xml" % self.target.output_filename.rsplit("-", 1)[0]
        outputfiles = sorted(
            glob.glob(str(Path(self.target.outputdir).joinpath(glob_pattern)))
        )
        return outputfiles


class CompleteRetry(RetryStrategy):
    """Execution strategy for suites with complete re-execution"""

    def __str__(self):
        return "Strategy: Complete"

    def _reparametrize(self):
        """Reparametrize the suite for the next attempt.

        Only incremental strategy needs to do something here."""
        pass


class IncrementalRetry(RetryStrategy):
    """Provides methods to re-execute suites incrementally (no test interdependency)"""

    def __str__(self):
        return "Strategy: Incremental"

    def _reparametrize(self):
        """Reparametrize the suite for the next attempt.

        The next attempt needs the XML file of the last attempt as input.
        From there it will read failed tests and re-execute them only."""
        # Chance for next try. Attempt gets increased, output files get bumped
        failed_xml = Path(self.target.outputdir).joinpath(self.target.output_xml)
        self.target.robot_params.update({"rerunfailed": str(failed_xml)})
        rerun_selection = self.target.config.get(
            "suitecfg.retry_failed.rerun_selection", asdict=True, default={}
        )
        self.target.robot_params.update(rerun_selection)



import loguru
from abc import ABC, abstractmethod
from datetime import datetime


# TODO: concurrent writes to the log file
# e.g. RCC -> RF


class AbstractLogger(ABC):
    def __init__(self, log_level):
        self.log_level = log_level
        if not getattr(self, "logger", None):
            self.logger = loguru.logger
            # Disable the logger completely if log level is not set
            if self.log_level is None:
                self.logger.disable()
            else:
                self.logger.remove()  # Remove default configuration
                self.add_logger()

    @abstractmethod
    def add_logger(self):
        pass

    def debug(self, message, *args, **kwargs):
        self.logger.debug(message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        self.logger.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.logger.warning(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.logger.error(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        self.logger.critical(message, *args, **kwargs)


class RobotmkLogger(AbstractLogger):
    def __init__(self, log_file_path, log_level="INFO"):
        self.log_file_path = log_file_path
        super().__init__(log_level)

    def add_logger(self):
        # Add the file sink
        self.logger.add(self.log_file_path, level=self.log_level)


# class JSONLogger(AbstractLogger):
#     """Logging into an internal message stack"""

#     def __init__(self, log_level="INFO"):
#         self.messages = []
#         self.logger.add(
#             {
#                 "sink": lambda msg: self.messages.append(
#                     self.message_to_dict(msg.record), log_level=log_level
#                 )
#             }
#         )

#     @staticmethod
#     def message_to_dict(record):
#         return {
#             **record,
#             "extra": {k: str(v) for k, v in record["extra"].items()},
#             "time": datetime.utcnow().isoformat() + "Z",
#         }



from pathlib import Path
from ..abstract import AbstractContext

from robotmk.config import Config, RobotmkConfigSchema
from robotmk.executor.scheduler import Scheduler
from robotmk.emitter import Emitter


class AgentContext(AbstractContext):
    def __init__(self):
        super().__init__()
        self.ymlschema = RobotmkConfigSchema

    @property
    def logger(self):
        if not self._logger:
            self._logger = RobotmkLogger(
                Path(self.config.common.logdir).joinpath("robotmk.log"),
                self.config.common.log_level,
            )
        return self._logger

    def load_config(self, defaults, **kwargs) -> None:
        """Load the config for agent context.

        Agent context can merge the config from
        - OS defaults
        - + environment variables
        - + YML file (default/custom = --yml)
        - + environment variables
        """
        # After the defaults are read...
        self.config.set_defaults(defaults)
        # ...lets first search for env vars which point to another location of the robotmk.yml!
        self.config.read_yml_cfg(path=kwargs["ymlfile"], must_exist=True)
        # In the end, environment variables can override everything
        self.config.read_cfg_vars(path=None)

        # TODO: validate later so that config can be dumped
        # self.config.validate(self._ymlschema)

    def refresh_config(self) -> bool:
        """Re-loads the config and returns True if it changed"""
        config_copy = copy.deepcopy(self.configdict)
        # re-initializes the config object
        super().__init__(envvar_prefix=self.envvar_prefix)
        config_changed = config_copy != self.configdict
        return config_changed

    def execute(self, *args, **kwargs):
        """Starts the scheduler."""
        self.executor = Scheduler(self.config)
        self.executor.run()
        pass

    def output(self):
        """Gathers the results of all scheduled suites and returns CMK agent output from it."""
        self.emitter = Emitter(self.config)
        self.emitter.run()



"""Run the log viewer web server.

Starts a local webserver to serve the execution logs of all Robot suites."""
import sys
import click
from robotmk.main import Robotmk, DEFAULTS


# use module docstring as help text
@click.command(help=__doc__)
@click.pass_context
def server(ctx):
    click.secho("server", fg="green")
    click.echo("Serving the Robotmk webinterface on http://localhost:8099 ...")
    # ctx.robotmk = Robotmk("server")
    # ctx.obj.config.set("common.context", "server")
    pass



import re
import os
from click.testing import CliRunner
import robotmk.cli.cli as cli

cwd = os.path.dirname(__file__)
robotmk_yml = os.path.join(cwd, "robotmk.yml")
robotmk_env = os.path.join(cwd, "robotmk.env")


def test_local_scheduler():
    """"""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["local", "--yml", robotmk_yml, "scheduler"])
    assert result.exit_code == 0


# test help message
def test_local_cli_help():
    """The help message should contain the three contexts and the help message itself."""
    runner = CliRunner()
    result = runner.invoke(cli.main, ["local", "--help"])
    assert result.exit_code == 0
    # assert "Robotmk CLI Interface." in result.output
    assert re.search(r"--yml TEXT", result.output, re.DOTALL)
    assert not re.search(r"--vars TEXT", result.output, re.DOTALL)
    assert re.search(r"Commands:.*output.*scheduler", result.output, re.DOTALL)


def test_local_cli_invalid_vars_option():
    """The --vars option should be invalid for the local context."""
    runner = CliRunner()
    result = runner.invoke(
        cli.main, ["local", "--vars", robotmk_env, "--yml", robotmk_yml, "scheduler"]
    )
    assert result.exit_code == 2
    assert "Error: No such option: --vars" in result.output



import time, sys, os
import json
from copy import deepcopy
from collections import defaultdict
from robotmk.main import Robotmk
from robotmk.context.suite.target.abstract import Target

from robotmk.context.suite.target.target_factory import TargetFactory

# from tabulate import tabulate


class Emitter:
    def __init__(self, config, *args, **kwargs):
        self.config = config

    def start_robotmk_process(self, run_env):
        cmd = "robotmk suite run".split(" ")
        result = subprocess.run(cmd, capture_output=True, env=run_env)
        # result = subprocess.run(["echo", "foo"], capture_output=True, env=run_env)
        stdout_str = result.stdout.decode("utf-8").splitlines()
        stderr_str = result.stderr.decode("utf-8").splitlines()
        result_dict = {
            "args": result.args,
            "returncode": result.returncode,
            "stdout": stdout_str,
            "stderr": stderr_str,
        }
        pass

    def prepare_environment(self, suiteuname) -> dict:
        run_env = os.environ.copy()
        added_settings = {
            "common.context": "suite",
            "common.suiteuname": suiteuname,
        }
        # run_env = basic config + added settings
        self.config.cfg_to_environment(self.config.configdict, environ=run_env)
        self.config.dotcfg_to_env(added_settings, environ=run_env)
        return run_env

    def run(self):
        """Iterates over all suites and produces agent output"""

        suites = self.config.get("suites")
        results = RMKResults()
        for suiteuname, suitecfg in suites:
            # TODO: handle logging
            self.config.set("common.suiteuname", suiteuname)
            target_ = TargetFactory(suiteuname, self.config, None).create()
            results.add(target_)
        print(results.get_results())


class RMKResults:
    def __init__(self):
        self.result_dict = defaultdict(list)

    def add(self, target: Target):
        self.result_dict[target.piggybackhost].append(target.output())

    def get_results(self):
        out = []
        host_results_copies = []
        robotmk_section_header = "<<<robotmk:sep(0)>>>"
        for (
            host,
            results,
        ) in self.result_dict.items():
            host_boundary = {
                True: [f"<<<<{host}>>>>", "<<<<>>>>"],
                False: ["", ""],
            }
            is_piggyback = host != "localhost"
            if is_piggyback:
                host_boundary = [f"<<<<{host}>>>>", "<<<<>>>>"]
            else:
                host_boundary = ["", ""]
            host_results = []
            for result in results:
                host_results.append(json.dumps(result, sort_keys=False, indent=4))
                if is_piggyback:
                    result_short = deepcopy(result)
                    del result_short["output"]
                    host_results_copies.append(result_short)

            host_results[0:0] = [robotmk_section_header]
            host_boundary[1:1] = host_results

            out += host_boundary
        if host_results_copies:  # if there are piggyback results
            out.append(robotmk_section_header)
            out += [
                json.dumps(result, sort_keys=False, indent=4)
                for result in host_results_copies
            ]
        return "\n".join(out)



"""This module provides a simple way to read configuration from different sources: 
- YML file
- variables file
- environment variables

There is a special order in which the sources are read:
- 1. OS defaults for each supported OS (Linux, Windows)
- 2. config from YML file, either
    - the default config file (robotmk.yml)
    OR
    - a custom config file (given as parameter --yml)
- 3. variables from 
    - variable file (given as parameter --vars)
    AND
    - environment variables (ROBOTMK_*)

| context       | yml    | vars |
| ---           | ---    | ---  |
| agent         | X      |      |
| specialagent  |        | X    |
| suite         | X      | X    |


"""

# TODO: add a mode to dump environment

import os
import re
import yaml
from mergedeep import merge, Strategy
from typing import Union
from collections import defaultdict
import json
import hashlib
from functools import wraps

# from collections import namedtuple
from pathlib import Path
from .yml import RobotmkConfigSchema

# TODO: add YML config validation
# ["%s:%s" % (v, os.environ[v]) for v in os.environ if v.startswith("ROBO")]


# def default_dict_constructor(loader, node):
#     return defaultdict(loader.construct_mapping(node))


# yaml.add_constructor(
#     "tag:yaml.org,2002:python/object/apply:collections.defaultdict",
#     default_dict_constructor,
# )


def add_path_prefix(method):
    """Decorator function for the get() method.

    If the given key is one of (logdir, tmpdir, resultdir, robotdir),
    AND the value does not start with a slash,
    THEN the returned value is added at the end of the prefix.
    A slash is added, if missing.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        value = method(self, *args, **kwargs)
        if args[0].split(".")[-1] in [
            "cfgdir",
            "logdir",
            "tmpdir",
            "resultdir",
            "robotdir",
        ] and not value.startswith(
            "/"
        ):  # paths which start with a slash are absolute
            prefix = self.get("common.path_prefix", None)
            if prefix:
                if not prefix.endswith("/"):
                    value = "/" + value
                value = prefix + value
            else:
                # no prefix given, so we use the value as is
                pass
        return value

    return wrapper


class Config:
    def __init__(self, envvar_prefix: str = "ROBOTMK", initdict: dict = {}):
        self.envvar_prefix = envvar_prefix
        if initdict:
            self.default_cfg = initdict
        else:
            self.default_cfg = {}
        self.yml_config = {}
        self.env_config = {}
        # this is a dict of all config values that were added by the user.
        # they are applied last and can overwrite any other config values.
        self.added_config = {}

    def __iter__(self):
        # TODO: iterator added afterwards - where is it useful?
        for key in self.configdict:
            yield key, self.get(key)

    def __translate_keys(self, name: str) -> list:
        """Translate a key name to a list of keys and respect shorthand 'suitecfg'."""
        keys = name.split(".")
        if keys[0] == "suitecfg":
            suitename = self.configdict["common"]["suiteuname"]
            keys[:1] = ["suites", suitename]
        return keys

    @add_path_prefix
    def get(self, name: str, default=None, asdict=False) -> str:
        """Get a value from the object with dot notation.

        Shorthands:
        - 'suitecfg' can be used for 'suites.<suiteuname>'.
        - 'basic_cfg' can be used for the basic config dict.

        Examples:
            cfg.get("common.logdir")
            cfg.get("suitecfg.run.rcc")
            cfg.get("basic_cfg.common.logdir") -> returns logdir value from the configuration
        """
        keys = self.__translate_keys(name)
        if keys[0] == "basic_cfg":
            m = self.basic_cfg
            keys = keys[1:]
        else:
            m = self.configdict
        # prev = self.configdict
        try:
            for key in keys:
                # prev = m
                # prev_k = key
                m = m.get(key, {})
            if m:  # non-empty value
                if type(m) is dict:
                    if asdict:
                        return m
                    else:
                        return Config(initdict=m)
                else:
                    return m
            else:  # empty dict
                return default

        except:
            return default

    def set(self, name: str, value: any) -> None:
        """Set a value in the object with dot notation.

        Shorthand 'suitecfg' can be used for 'suites.<suiteuname>'.

        Example:
            cfg.set("common.logdir", "/foo/log")
            cfg.set("suitecfg.run.rcc", False)
        """
        keys = self.__translate_keys(name)
        cur_dict = self.added_config
        for key in keys[:-1]:
            if not key in cur_dict:
                cur_dict[key] = {}
            cur_dict = cur_dict[key]

        cur_dict[keys[-1]] = value

    def asdict(self):
        """Returns the config as a dict."""
        return self.configdict

    def suite_cfghash(self, suiteuname) -> str:
        """Returns a hash of the common + suite config (passed as arg) to identify a possible change.

        Used in the scheduler and sequencer."""
        common_cfg = {"common": self.get("common").asdict()}
        suite_cfg = {"suites": self.get("suites.%s" % suiteuname).asdict()}
        cfg = merge({}, common_cfg, suite_cfg)
        return hashlib.sha256(json.dumps(cfg).encode("utf-8")).hexdigest()

    @property
    def configdict(self):
        """This property represents all config dicts merged.

        It consist of the following dicts:
        - basic_cfg = predefiend from start
            - default_cfg
            - yml_config
            - env_config
        - added_config = added at runtime

        """

        # https://mergedeep.readthedocs.io/en/latest/#merge-strategies
        # Collection values are merged additively. Others get replaced.
        # The last/rightmost dictionary has the highest priority.
        return merge(
            {},
            self.basic_cfg,  # Config set by OS default values, YML file, environment variables and/or variables file
            self.added_config,  # Config changed/added at runtime (config.set() method)
            strategy=Strategy.ADDITIVE,
        )

    @configdict.setter
    def configdict(self, cfg: dict):
        """This setter allows to write a whole new config dict. It is used e.g. when the scheduler creates
        job object for each suite."""
        self.default_cfg = cfg

    @property
    def basic_cfg(self):
        """Returns the basic config dict."""
        return merge(
            {},
            self.default_cfg,  # Config set by OS default values
            self.yml_config,  # Config loaded from YML file
            self.env_config,  # Config loaded from environment variables and/or variables file
            strategy=Strategy.ADDITIVE,
        )

    # 1. Defaults (common/OS specific)
    def set_defaults(self, os_defaults: dict = None) -> None:
        """Sets the defaults for the current OS."""
        self.default_cfg["common"] = {}
        if os_defaults:
            self.default_cfg["common"].update(os_defaults["common"])
        if os.name in os_defaults:
            self.default_cfg["common"].update(os_defaults[os.name])

    # 2. YML
    def read_yml_cfg(self, path=None, must_exist=True):
        """Reads a YML config, either default or custom."""
        if path is None:
            # Linux default: /etc/check_mk/robotmk.yml
            # Windows default: C:\Program Data\check_mk\agent\config\robotmk.yml
            ymlfile = Path(self.get("common.cfgdir")) / self.get("common.robotmk_yml")
        else:
            ymlfile = Path(path)
            # a custom file path should always exist
            must_exist = True
        if must_exist and not ymlfile.exists():
            raise FileNotFoundError(f"YML config file not found: {ymlfile}")
        else:
            # try to read the file
            config = {}
            try:
                with open(ymlfile, "r") as f:
                    config = yaml.load(f, Loader=yaml.FullLoader)
            except Exception as e:
                raise e

            self.yml_config = config

    # 3. variables (env AND! file)
    def read_cfg_vars(self, path=None):
        """Read ROBOTMK variables from file and/or environment.

        Environment vars have precedence over file vars."""

        filevars = self._filevar2dict(path)
        envvars = self._envvar2dict()
        # a dict with still flat var names
        vars = merge(filevars, envvars)
        # convert flat vars to nested dicts
        vdict = defaultdict(dict)
        for k, v in vars.items():
            # iterate over each variable and convert it to a nested data structure
            vdict = self._var2cfg(k, v, vdict)
        self.env_config = merge({}, self.env_config, vdict, strategy=Strategy.ADDITIVE)

    def _envvar2dict(self) -> dict:
        """Returns all environment variables starting with the ROBOTMK prefix.

        Example:
        {"ROBOTMK_foo_bar": "baz",
         "ROBOTMK_foo_baz": "bar"}
        """
        vardict = {}
        for k, v in os.environ.items():
            if k.startswith(self.envvar_prefix):
                # check if the value is a boolean and convert it to a boolean
                if v.lower() in ("true", "false"):
                    v = v.lower() == "true"
                vardict[k] = v
        return vardict

    def _filevar2dict(self, file) -> dict:
        """Returns all variables from a given file (strips 'set' and 'export' statements).

        Example:
        {"ROBOTMK_foo_bar": "baz",
         "ROBOTMK_foo_baz": "bar"}
        """
        r = {}
        if file:
            try:
                with open(file, "r") as f:
                    for line in f:
                        line = line.strip()
                        # Ignore empty lines and lines starting with "#" (comments)
                        if line.strip() and not line.strip().startswith("#"):
                            if line.startswith("export ") or line.startswith("set "):
                                line = line.partition(" ")[2]
                            # Split each line into a key-value pair
                            key, value = line.strip().split("=")
                            if key.startswith(self.envvar_prefix):
                                r[key] = value
            except Exception as e:
                raise FileNotFoundError(f"Could not read environment file: {file}")
        return r

    def validate(self, schema: RobotmkConfigSchema):
        """Validates the whole config according to the given context schema."""

        schema = RobotmkConfigSchema(self.configdict)
        if not schema.validate():
            raise ValueError(f"Config is invalid: {schema.error}")

    @staticmethod
    def partition_at_digit(s):
        """Divides the string at the first digit and returns a list of the
        left part, the digit and the right part.

        The digit indicates the index in the list."""
        m = re.search("_\d+", s)
        if m:
            return s[: m.start()], int(s[m.start() + 1 : m.end()]), s[m.end() + 1 :]
        else:
            return [s, None, None]

    @staticmethod
    def split_varstring(s):
        """Helper function to reconstruct the nested key path of a dict from
        a varname. The keys are separated by "_", a double underscore protects the string from splitting.
        """
        keys = []
        starti = 0
        i = 0
        while i < len(s):
            poschar = s[i]
            if poschar == "_" or i == len(s) - 1:
                if len(s) > i + 1 and s[i + 1] == "_":
                    # not at and and double underscore in front of us!
                    # skip next underscore and continue until next SINGLE underscore
                    i += 2
                    continue
                else:
                    # Single underscore in front or last piece; add current piece to list
                    # (replace __ by _) and start a new one
                    if len(s) > i + 1:
                        # last piece
                        last_single_usc_index = i
                    else:
                        # not last piece
                        last_single_usc_index = i + 1
                    piece = s[starti:last_single_usc_index].replace("__", "_")
                    keys.append(piece)
                    starti = i + 1
            else:
                # Add the current character to the current piece
                pass
            i += 1
        return keys

    def uscore_str2dict(string, value, vdict=None):
        """Converts a variable string to a nested dict.

        Digit inside the string indicate the index in the list."""
        # curdict is a moving reference to the current dict
        cur_dict = vdict

        (s_left, s_list_index, s_right) = Config.partition_at_digit(string)
        # list of keys "left" of the list index (if any)
        left_keys = Config.split_varstring(s_left)
        for ki, key in enumerate(left_keys):
            # descend now key by key until we reach the leaf key. cur_dict gets always
            # moved forward to have a reference on the key we want to change.
            if ki < len(left_keys) - 1:
                if not key in cur_dict:
                    cur_dict[key] = defaultdict(dict)
                cur_dict = cur_dict[key]
            else:
                # we have reached the leaf key of left side, just before the index.
                if not s_list_index is None:  # list index
                    if (
                        not key in cur_dict
                    ):  # key not present, create all (empty) list items
                        cur_dict[key] = [defaultdict(dict)] * (s_list_index + 1)
                    else:
                        if (
                            len(cur_dict[key]) < s_list_index + 1
                        ):  # key present, but not enough list items
                            cur_dict[key] = cur_dict[key] + [defaultdict(dict)] * (
                                s_list_index + 1 - len(cur_dict[key])
                            )
                    if not s_right:  # list entry without subkeys
                        cur_dict[key][s_list_index] = value
                    else:
                        # dict inside the list!
                        # call the function recursively to descend into the dict. We also
                        # pass the current list item as cur_dict, so that the function can
                        # modify it.
                        subdict = Config.uscore_str2dict(
                            s_right, value, cur_dict[key][s_list_index]
                        )
                        # assign the modified subdict to the list item
                        cur_dict[key][s_list_index] = subdict
                else:
                    # simple value assignment. Make sure that values of numbers are not stored as strings.
                    if isinstance(value, str) and value.isdigit():
                        value = int(value)
                    cur_dict[key] = value

        return vdict

    def _var2cfg(self, o_varname, value, vdict) -> dict:
        """Helper function to convert a variable to a dict/list/value entry and assigns the value.

        Example:
            - ROBOTMK_foo_bar_0_baz = "bar"
                will be converted to:
                {"foo": {"bar": [{"baz": "bar"}]}}
            - ROBOTMK_foo__bar_0_baz = "bar
                will be converted to:
                {"foo_bar": [{"baz": "bar"}]}
        """

        # Remove the ROBOTMK_ prefix
        varname = o_varname.replace(self.envvar_prefix + "_", "")
        Config.uscore_str2dict(varname, value, vdict)
        return vdict

    def dotcfg_to_env(self, dotstrings: dict, environ: dict):
        """Converts a dict of dotconfigs to an environment variable and assign it to environ.

        This does not affect the current config at all; it just uses a temporary config object.
        Example:
            {"common.logdir": "foo",
             "common.loglevel": "debug"}
        """

        tmp_cfg = Config()
        # As we are using a temp config object, the suiteuname must be set manually to access the
        # correct section with the shorthand "suitecfg"
        tmp_cfg.set("common.suiteuname", self.get("common.suiteuname"))
        for k, v in dotstrings.items():
            tmp_cfg.set(k, v)
        self.cfg_to_environment(tmp_cfg.configdict, environ=environ)

    def cfg_to_environment(self, d, prefix="", environ=None):
        """Converts a given dict to environment variables.

        If environ is given, the environment variables are added to the given dict.
        Otherwise the environment variables are added to the current environment.

        The conversion rules are:
        - there is no case conversion
        - underscores within key names are replaced by double underscores
        - the prefix is added to the environment variable name
        - dicts are converted to nested environment variables
        - lists get a number appended to their key name"""

        if isinstance(d, dict):  # DICT conversion
            for key, value in d.items():
                safe_key = key.replace("_", "__")
                new_prefix = f"{prefix}_{safe_key}"
                self.cfg_to_environment(value, prefix=new_prefix, environ=environ)
        elif isinstance(d, list):  # LIST conversion
            for i, item in enumerate(d):
                new_prefix = f"{prefix}_{i}"
                self.cfg_to_environment(item, prefix=new_prefix, environ=environ)
        else:  # VALUE conversion
            varname = f"{self.envvar_prefix}{prefix}"
            if isinstance(d, bool):
                # convert bools to lower case strings
                d = str(d).lower()
            print(f"{varname}={d}")
            if environ is None:
                os.environ[varname] = str(d)
            else:
                environ[varname] = str(d)

    def to_yml(self, file=None) -> Union[str, None]:
        """Dumps the config to a file or returns it."""
        if file:
            try:
                with open(file, "w") as f:
                    yaml.dump(self.configdict, f)
            except Exception as e:
                print(f"Could not write to file {file}: {e}")
                return None
        else:
            return yaml.dump(
                self.configdict, sort_keys=False, indent=4, default_flow_style=False
            )



import click
import json
import pprint

"""
@click.group(<name>) creates a command that instantiates a group class
a group is intended to be a set of related commands
@click.argument(<argument name>) tells us that we will be passing an argument
and referring to that argument in the function by the name we pass it
@click.pass_context tells the group command that we're going to be using
the context, the context is not visible to the command unless we pass this
 
In our example we'll name our group "cli"
"""


@click.group("cli")
@click.pass_context
@click.argument("document")
def cli(ctx, document):
    """An example CLI for interfacing with a document"""
    _stream = open(document)
    _dict = json.load(_stream)
    _stream.close()
    ctx.obj = _dict


@cli.command("check_context_object")
@click.pass_context
def check_context(ctx):
    pprint.pprint(type(ctx.obj))


pass_dict = click.make_pass_decorator(dict)


@cli.command("get_keys")
@pass_dict
def get_keys(_dict):
    keys = list(_dict.keys())
    click.secho("The keys in our dictionary are", fg="green")
    click.echo(click.style(keys, fg="blue"))


@cli.command("get_key")
@click.argument("key")
@click.pass_context
def get_key(ctx, key):
    pprint.pprint(ctx.obj[key])


@click.option("-d", "--download", is_flag=True, help="Download the results")
@click.option("-k", "--key", help="The key to search for")
@click.pass_context
@cli.command("get_results")
def get_results(ctx, download: bool, key: str):
    results = ctx.obj["results"]
    if key is not None:
        result = {}
        for entry in results:
            if key in entry:
                if key in result:
                    result[key] += entry[key]
                else:
                    result[key] = entry[key]
        results = result
    if download:
        if key is not None:
            filename = key + ".json"
        else:
            filename = "results.json"
        with open(filename, "w") as w:
            w.write(json.dumps(results))
        print("File saved to", filename)
    else:
        pprint.pprint(results)


def main():
    cli(prog_name="cli")


if __name__ == "__main__":
    main()



from robotmk.main import Robotmk, DEFAULTS
import pytest
import os


def test_robotmk_no_context_in_env():
    """Test that Robotmk throws an error if created without any context in environment."""
    match_msg = (
        "No context given on CLI or set by environment variable ROBOTMK_common_context."
    )
    with pytest.raises(ValueError, match=match_msg):
        robotmk = Robotmk()


def test_robotmk_local_context():
    """Test that Robotmk has agent context when set in environment."""
    os.environ["ROBOTMK_common_context"] = "agent"
    robotmk = Robotmk()
    assert robotmk.config.get("common.context") == "agent"


def test_robotmk_suite_context():
    """Test that Robotmk has suite context when set in environment."""
    os.environ["ROBOTMK_common_context"] = "suite"
    robotmk = Robotmk()
    assert robotmk.config.get("common.context") == "suite"


def test_robotmk_specialagent_context():
    """Test that Robotmk has specialagent context when set in environment."""
    os.environ["ROBOTMK_common_context"] = "specialagent"
    robotmk = Robotmk()
    assert robotmk.config.get("common.context") == "specialagent"



import time, sys, os

from apscheduler.schedulers.background import BackgroundScheduler

# from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.executors.pool import ProcessPoolExecutor
from .abstract import AbstractExecutor
from robotmk.main import Robotmk
import subprocess
from tabulate import tabulate


class Scheduler(AbstractExecutor):
    def __init__(self, config, *args, **kwargs):
        # TODO: Max. number of prcesses must be lower than the number of CPUs
        super().__init__(config)
        self.scheduler = BackgroundScheduler(
            executors={"mydefault": ProcessPoolExecutor(6)}
        )
        self.suitecfg_hashes = {}
        # self.scheduler = BlockingScheduler(
        #     executors={"mydefault": ProcessPoolExecutor(6)}
        # )

    def start_robotmk_process(self, run_env):
        cmd = "robotmk suite run".split(" ")
        result = subprocess.run(cmd, capture_output=True, env=run_env)
        # result = subprocess.run(["echo", "foo"], capture_output=True, env=run_env)
        stdout_str = result.stdout.decode("utf-8").splitlines()
        stderr_str = result.stderr.decode("utf-8").splitlines()
        result_dict = {
            "args": result.args,
            "returncode": result.returncode,
            "stdout": stdout_str,
            "stderr": stderr_str,
        }
        pass

    def prepare_environment(self, suiteuname) -> dict:
        run_env = os.environ.copy()
        added_settings = {
            "common.context": "suite",
            "common.suiteuname": suiteuname,
        }
        # run_env = basic config + added settings
        self.config.cfg_to_environment(self.config.configdict, environ=run_env)
        self.config.dotcfg_to_env(added_settings, environ=run_env)
        return run_env

    def schedule_jobs(self):
        """Updates the scheduler with new jobs and removes old ones"""
        suites = self.config.get("suites")
        current_jobs = set(self.scheduler.get_jobs())
        # remove jobs that are no longer in the config
        for job in current_jobs - set(suites.asdict().keys()):
            self.scheduler.remove_job(job.id)
        # schedule new/changed jobs
        for suiteuname, suitecfg in suites:
            hash = self.config.suite_cfghash(suiteuname)
            if hash == self.suitecfg_hashes.get(suiteuname, {}):
                # The suite config has not changed, so we can skip it
                continue
            else:
                run_env = self.prepare_environment(suiteuname)
                interval = suitecfg.get("scheduling.interval")

                self.scheduler.add_job(
                    self.start_robotmk_process,
                    args=[run_env],
                    trigger="interval",
                    id=suiteuname,
                    seconds=interval,
                    replace_existing=True,
                    # args=[v],
                    max_instances=1,
                )
                self.suitecfg_hashes[suiteuname] = hash

    def run(self):
        """Start the scheduler and update the jobs every 5 seconds"""

        self.schedule_jobs()
        # self.scheduler.add_listener(self.log)
        self.scheduler.start()
        while True:
            # update the jobs every 5 seconds
            time.sleep(5)

            jobs = self.scheduler.get_jobs(jobstore=None)
            table = []
            for job in jobs:
                table.append(
                    [
                        job.id,
                        job.name,
                        # job.args,
                        job.trigger,
                        job.pending,
                        job.next_run_time,
                        job.trigger.interval.total_seconds(),
                    ]
                )
            # print current time
            print(time.strftime("%H:%M:%S", time.localtime()))
            print(
                tabulate(
                    table,
                    headers=[
                        "id",
                        "name",
                        "trigger",
                        "pending",
                        "next_run at",
                        "interval",
                    ],
                )
            )
            print("\n")
